#include <cstdio>
#include <stdlib.h>

typedef unsigned long UInt64;
typedef long Int64;
typedef UInt64 SizeType;

typedef UInt64 Unsigned;
typedef Int64 Integer;
typedef bool Boolean;
typedef float Float;
typedef double Double;

typedef unsigned int bin_id;
typedef const bin_id* binset_const_iter;

namespace Endeca { namespace Jit {
class CursorNode {
public:
	void * getConstructorArgValue(int position) { return NULL; }
	CursorNode * getChildPtr(int position) { return NULL; }
};
class Double_DLCursorWrapper {
public:
	void advance() {}
	void seek(unsigned current) {}
	bool isValid() { return false; }
	UInt64 getUnsigned(unsigned col) { return 0; }
	double getDouble(unsigned col) { return 0.0; }
};
class Integer_DLCursorWrapper {
public:
	void advance() {}
	void seek(unsigned current) {}
	bool isValid() { return false; }
	Int64 getInteger(unsigned col) { return 0; }
	UInt64 getUnsigned(unsigned col) { return 0.0; }
};
class Unsigned_DLCursorWrapper {
public:
	void advance() {}
	void seek(unsigned current) {}
	bool isValid() {return false;}
	UInt64 getUnsigned(unsigned col) { return 0;}
};
class StringId_DLCursorWrapper {
public:
	void advance(){}
	void seek(unsigned current){}
	bool isValid(){return false;}
	UInt64 getUnsigned(unsigned col){return 0;}
};
class CategoryId_DLCursorWrapper {
public:
	void advance(){}
	bool isValid(){return false;}
	void seek(unsigned current){}
	UInt64 getUnsigned(unsigned col){return 0;}
};
class UnsignedTest_DLCursorWrapper {
public:
	void advance(){}
	void seek(unsigned current){}
	bool isValid(){return false;}
	UInt64 getUnsigned(unsigned col){return 0;}
};
class IntegerTest_DLCursorWrapper {
public:
	void advance(){}
	void seek(unsigned current){}
	bool isValid(){return false;}
	Int64 getInteger(unsigned col){return 0;}
	UInt64 getUnsigned(unsigned col){return 0;}
};
class Double_RectangularCursorWrapper {
public:
	UInt64 getUnsigned(unsigned col){return 0;}
	double getDouble(unsigned col){return 0.0;}
	bool sync(binset_const_iter current){return false;}
};
class Integer_RectangularCursorWrapper {
public:
	Int64 getInteger(unsigned col){return 0;}
	UInt64 getUnsigned(unsigned col){return 0;}
	bool sync(binset_const_iter current){return false;}
};
class Unsigned_RectangularCursorWrapper {
public:
	UInt64 getUnsigned(unsigned col){return 0;}
	bool sync(binset_const_iter current){return false;}
};
class StringId_RectangularCursorWrapper {
public:
	UInt64 getUnsigned(unsigned col){return 0;}
	bool sync(binset_const_iter current){return false;}
};
class CategoryId_RectangularCursorWrapper {
public:
	UInt64 getUnsigned(unsigned col){return 0;}
	bool sync(binset_const_iter current){return false;}
};
class UnsignedTest_RectangularCursorWrapper {
public:
	UInt64 getUnsigned(unsigned col){return 0;}
	bool sync(binset_const_iter current){return false;}
};
}
class ColumnNumber {
	unsigned ** key;
public:
	unsigned value;
};

namespace XQuery{ namespace Relational { namespace Evaluation {
class Table {
public:
	SizeType nRows;
	void** starts_begin;
	void** starts_finish;
	void** starts_end_of_storage;

	class TableWrapper {
	public:
		Unsigned getUnsigned(unsigned row, unsigned col) const {return 0;}
		Boolean getBoolean(unsigned row, unsigned col) const {return false;}
		Integer getInteger(unsigned row, unsigned col) const {return 0;}
		Float getFloat(unsigned row, unsigned col) const {return 0.0;}
		Double getDouble(unsigned row, unsigned col) const {return 0.0;}
	};
};

class Sink {
	char virtual_function_table[8];
public:
	SizeType baseRow;
	const ColumnNumber baseCol;
	SizeType flushThreshold;
	unsigned width;
	void * functorListHead;
	SizeType curRow;
	Table *buffer;
};

class SinkWrapper {
public:
	Sink *sink;

	void addRow(){}
	void resize(SizeType atLeast){}
	void flushRow(){}
	void flush(){}
	void pushUnsigned(unsigned col, Unsigned value){}
	void pushBoolean(unsigned col, Boolean value){}
	void pushInteger(unsigned col, Integer value){}
	void pushFloat(unsigned col, Float value){}
	void pushDouble(unsigned col, Double value){}
};

class RowExprCursor {
public:
	class Function {
	public:
		virtual ~Function() {}
		virtual void operator()(
			void* ret,
			void* left = NULL,
			void* middle = NULL,
			void* right = NULL
		){}
	};
	};
} } } }

namespace XRE = Endeca::XQuery::Relational::Evaluation;

template<typename Comp> inline void
push_heap(
	unsigned* first, 
	unsigned holeIndex,
	unsigned topIndex, 
	unsigned value, 
	Comp comp
) {
	unsigned parent = (holeIndex - 1 ) / 2;
	while ( holeIndex > topIndex && comp(*(first + parent), value)) {
		*(first + holeIndex) = *(first + parent);
		holeIndex = parent;
		parent = (holeIndex - 1 ) / 2;
	}
	*(first + holeIndex) = value;
}

template<typename Comp> inline void
push_heap(
	unsigned* first,
	unsigned* last,
	Comp comp
) {
	push_heap(first, unsigned((last - first) - 1), unsigned(0), unsigned(*(last - 1)), comp);	
}

template<typename Comp> inline void
adjust_heap(
	unsigned* first,
	unsigned holeIndex,
	unsigned length,
	unsigned value,
	Comp comp
) {
	const unsigned topIndex = holeIndex;
	unsigned secondChild = 2 * holeIndex + 2;
	while ( secondChild < length ) {
		if ( comp(*(first + secondChild), *(first + (secondChild-1)))) {
			secondChild--;
		}
		*(first + holeIndex) = *(first + secondChild);
		holeIndex = secondChild;
		secondChild = 2 * (secondChild + 1);
	}
	if ( secondChild == length ) {
		*(first + holeIndex) = *(first + (secondChild-1));
		holeIndex = secondChild - 1;
	}
	push_heap(first, holeIndex, topIndex, value, comp);
}

template<typename Comp> inline void
pop_heap(
	unsigned* first,
	unsigned* last,
	unsigned* result,
	unsigned value,
	Comp comp	
) {
	*result = *first;
	adjust_heap(first, unsigned(0), unsigned(last - first), value, comp);
}

template<typename Comp> inline void
pop_heap(
	unsigned* first,
	unsigned* last,
	Comp comp
) {
	pop_heap(first, last - 1, last - 1, unsigned(*(last-1)), comp);	
}

template<typename Comp> inline void
make_heap(
	unsigned* first,
	unsigned* last,
	Comp comp
) {
	if ( last - first < 2 ) return;

	const unsigned len = last - first;
	unsigned parent = (len - 2) / 2;
	while ( true ) {
		adjust_heap(first, parent, len, unsigned(*(first+parent)), comp);
		if ( parent == 0 ) {
			return;
		}
		parent--;
	}
}

class TableCursorIntegerInteger {
	XRE::Table::TableWrapper *table;
	unsigned current;
	unsigned stop;
public:
	TableCursorIntegerInteger(Endeca::Jit::CursorNode * node) :
		table((XRE::Table::TableWrapper *)node->getConstructorArgValue(0)),
		current(*(unsigned*)node->getConstructorArgValue(1)),
		stop(*(unsigned*)node->getConstructorArgValue(2))
	{ }

	void advance() {
		current++;
	}

	bool isValid() const {
		return (current < stop);
	}

	Integer getItem0() const {
		return table->getInteger(current, 0);
	}

	Integer getItem1() const {
		return table->getInteger(current, 1);
	}

	void seek(const Integer target) {
		if (current >= stop || getItem0() >= target) {
			return;
		}

		int j;
		for(j=1; current+j < stop && table->getInteger(current + j, 0) < target; j <<= 1) current += j;
		while(j >>= 1) if (current+j < stop && table->getInteger(current + j, 0) < target) current += j;

		++current;
	}

	void seekPast(const Integer target) {
		seek(target);
		advanceToNextChunk();
	}
	bool isLastRowOfChunk() const {
		if (current + 1 >= stop) return true;
		return table->getInteger(current, 0) != table->getInteger(current + 1, 0);
	}

	void advanceToNextChunk() {
		while (!isLastRowOfChunk()) advance();
		advance();
	}
};



struct MergeJoinCursor_0_Data_Struct {
	Integer item1;
};

class MergeJoinCursor_0_Vector {
	unsigned long size_, capacity;
	MergeJoinCursor_0_Data_Struct *data;

public:
	MergeJoinCursor_0_Vector() :
		size_(0),
		capacity(10),
		data((MergeJoinCursor_0_Data_Struct *)malloc(capacity * sizeof(MergeJoinCursor_0_Data_Struct)))
	{ }

	~MergeJoinCursor_0_Vector() {
		free(data);
	}

	void push_back(MergeJoinCursor_0_Data_Struct item) {
		if (size_ >= capacity) {
			capacity *= 2;
			data = (MergeJoinCursor_0_Data_Struct *)realloc(data, capacity * sizeof(MergeJoinCursor_0_Data_Struct));
		}

		data[size_] = item;
		size_++;
	}

	void clear() {
		size_ = 0;
	}

	unsigned long size() const {
		return size_;
	}

	MergeJoinCursor_0_Data_Struct operator[](unsigned long i) const {
((i < size_) ? ((void)0) : (fprintf(stderr, "assertion i < size_ failed at xquery/relational/evaluation/cursors/MergeJoinCursor.cpp:287\n"), abort()));
		return data[i];
	}
};

class MergeJoinCursor_0 {
	TableCursorIntegerInteger left;
	TableCursorIntegerInteger right;

	unsigned long rightBufferIndex;
	MergeJoinCursor_0_Vector rightBuffer;

public:
	MergeJoinCursor_0(Endeca::Jit::CursorNode * mjcNode) : 
	left(mjcNode->getChildPtr(0)), 
	right(mjcNode->getChildPtr(1)), 
	rightBufferIndex(0)
	{
		startNextChunk();
	}

	void startNextChunk() {
		if (matchPrefix()) fillRightBuffer();
	}

	bool matchPrefix() {
		bool leftIsValid = false;
		bool rightIsValid = false;

		while ((leftIsValid || left.isValid()) && (rightIsValid || right.isValid())) {
			leftIsValid = rightIsValid = true;
			Integer leftKey = left.getItem0();
			Integer rightKey = right.getItem0();

			if (leftKey == rightKey) {
				return true;
			} else if (leftKey < rightKey) {
				left.seek(rightKey);
				leftIsValid = false;
			} else {
				right.seek(leftKey);
				rightIsValid = false;
			}
		}
		return false;
	}

	void fillRightBuffer() {
		rightBuffer.clear();
		while (1) {
			MergeJoinCursor_0_Data_Struct row;
			row.item1 = right.getItem1();
			rightBuffer.push_back(row);

			if (right.isLastRowOfChunk()) {
				break;
			} else {
				right.advance();
			}
		}
	}

	bool isValid() const {
		return (left.isValid() && right.isValid());
	}

	bool isLastRowOfChunk() {
		return (left.isLastRowOfChunk() && (rightBufferIndex + 1) >= rightBuffer.size());
	}

	void advance() {
		rightBufferIndex++;

		if (rightBufferIndex >= rightBuffer.size()) {
			rightBufferIndex = 0;

			if (left.isLastRowOfChunk()) {
				left.advance();
				right.advance();

				startNextChunk();
			} else {
				left.advance();
			}
		}
	}

	void advanceToNextChunk() {
		left.advanceToNextChunk();
		rightBufferIndex = 0;
		startNextChunk();
	}

	void seek(const Integer target) {
		left.seek(target);
		rightBufferIndex = 0;
		startNextChunk();
	}

	void seekPast(const Integer target) {
		left.seekPast(target);
		rightBufferIndex = 0;
		startNextChunk();
	}

	Integer getItem0() {
		return left.getItem0();
	}

	Integer getItem1() {
		return left.getItem1();
	}

	Integer getItem3() {
		return rightBuffer[rightBufferIndex].item1;
	}

};

class AllCursor {
	MergeJoinCursor_0 child;
	bool internalValid;
	bool childValid;
	Integer col1;
	Integer col3;

public:
	AllCursor(Endeca::Jit::CursorNode* allNode) : 
		child(allNode->getChildPtr(0)), internalValid(true), childValid(child.isValid())
	{
		advance();
	}

	void updateBuffer() {
((childValid) ? ((void)0) : (fprintf(stderr, "assertion childValid failed at xquery/relational/evaluation/cursors/AllCursor.cpp:124\n"), abort()));
((internalValid) ? ((void)0) : (fprintf(stderr, "assertion internalValid failed at xquery/relational/evaluation/cursors/AllCursor.cpp:125\n"), abort()));
			col1 = child.getItem1();
			col3 = child.getItem3();
			child.advance();
			childValid = child.isValid();
	}

	bool isValid() const {
((! (childValid && ! internalValid)) ? ((void)0) : (fprintf(stderr, "assertion ! (childValid && ! internalValid) failed at xquery/relational/evaluation/cursors/AllCursor.cpp:139\n"), abort()));
		return internalValid;
	}

	bool isLastRowOfChunk() const {
((internalValid) ? ((void)0) : (fprintf(stderr, "assertion internalValid failed at xquery/relational/evaluation/cursors/AllCursor.cpp:144\n"), abort()));
		return !childValid;
	}

	void advance() {
((internalValid) ? ((void)0) : (fprintf(stderr, "assertion internalValid failed at xquery/relational/evaluation/cursors/AllCursor.cpp:149\n"), abort()));
		if (childValid) {
			updateBuffer();
		} else {
			internalValid = false;
		}
	}

	void advanceToNextChunk() {
((internalValid) ? ((void)0) : (fprintf(stderr, "assertion internalValid failed at xquery/relational/evaluation/cursors/AllCursor.cpp:158\n"), abort()));
		internalValid = false;
	}

	Integer getItem1() {
((internalValid) ? ((void)0) : (fprintf(stderr, "assertion internalValid failed at xquery/relational/evaluation/cursors/AllCursor.cpp:196\n"), abort()));
		return col1;
	}

	Integer getItem3() {
((internalValid) ? ((void)0) : (fprintf(stderr, "assertion internalValid failed at xquery/relational/evaluation/cursors/AllCursor.cpp:196\n"), abort()));
		return col3;
	}

};


class Shim_AllCursor {
	AllCursor child;

public:
	Shim_AllCursor(Endeca::Jit::CursorNode * childNode) :
		child(childNode)
	{
	}

	bool fetch(XRE::SinkWrapper *sink_wrapper) {
((sizeof(Int64) == 8) ? ((void)0) : (fprintf(stderr, "assertion sizeof(Int64) == 8 failed at xquery/relational/evaluation/cursors/Shim.cpp:651\n"), abort()));
((sizeof(UInt64) == 8) ? ((void)0) : (fprintf(stderr, "assertion sizeof(UInt64) == 8 failed at xquery/relational/evaluation/cursors/Shim.cpp:652\n"), abort()));
		XRE::Sink *sink = sink_wrapper->sink;
		void ** const starts = sink->buffer->starts_begin;
		const unsigned baseCol = sink->baseCol.value;
		bool gotARow = false;

		while (child.isValid()) {
			if (sink->buffer->nRows < sink->curRow + 1) sink_wrapper->resize(sink->curRow + 1);

			((Integer *) starts[1 + baseCol])[sink->curRow] = child.getItem1();
			((Integer *) starts[3 + baseCol])[sink->curRow] = child.getItem3();

			sink->curRow++;
			if (sink->curRow - sink->baseRow >= sink->flushThreshold) sink_wrapper->flush();

			gotARow = true;
			bool done = child.isLastRowOfChunk();

			child.advance();

			if (done) break;
		}

		return gotARow;
	}
};

Shim_AllCursor * shim;
extern "C" void create(Endeca::Jit::CursorNode* childNode) {
	shim = new Shim_AllCursor(childNode);
}
extern "C" bool fetch(XRE::SinkWrapper *sink_wrapper) { 
	//printf("shim jit: fetching\n");
	return shim->fetch(sink_wrapper);
}


