#include <iostream>
#include <vector>

#include "llvm/LLVMContext.h"
#include "llvm/Target/TargetSelect.h"
#include "llvm/Module.h"
#include "llvm/ExecutionEngine/ExecutionEngine.h"
#include "llvm/ExecutionEngine/JIT.h"
#include "llvm/Support/ErrorHandling.h"
#include "llvm/Support/ManagedStatic.h"
#include "llvm/Target/TargetSelect.h"

#include "clang/Frontend/CompilerInstance.h"
#include "clang/Frontend/TextDiagnosticBuffer.h"
#include "clang/Frontend/FrontendDiagnostic.h"
#include "clang/CodeGen/CodeGenAction.h"

#define THREADS 100

void* go(void* arg)
{
	clang::CompilerInstance CI;
	if ( !CI.hasLLVMContext() ) CI.setLLVMContext(new llvm::LLVMContext);

	clang::TextDiagnosticBuffer *DiagBuffer = new clang::TextDiagnosticBuffer();
	clang::Diagnostic Diags(DiagBuffer);

    // Create the compiler arguments
    llvm::SmallVector<const char*, 32> args;
    args.push_back("-cc1");
    args.push_back("-std=c++98");
    args.push_back("-O3");
	args.push_back("code.cpp");
    //args.push_back("-g");
    //args.push_back(packagedCode);

    // Create a compiler invocation
    clang::CompilerInvocation *invocation = new clang::CompilerInvocation();
    clang::CompilerInvocation::CreateFromArgs(*invocation, args.begin(), args.end(), Diags);
    CI.setInvocation(invocation); // takes ownership of the invocation

    CI.createDiagnostics(args.size(), const_cast<char**>(args.begin()));
    DiagBuffer->FlushDiagnostics(CI.getDiagnostics());

    // Create and execute the frontend to generate an LLVM bitcode module.
    llvm::OwningPtr<clang::CodeGenAction> Act(new clang::EmitLLVMOnlyAction());
    if (!CI.ExecuteAction(*Act)) {
        std::cerr << "Failed to compile code:" << std::endl;
        std::cerr << "code.cpp" << std::endl;
		exit(-1);
    }

	llvm::Module* compiledModule = Act->takeModule();

	llvm::ExecutionEngine* EE = NULL;
	std::string err;
	{
		EE = llvm::EngineBuilder(compiledModule).setOptLevel(llvm::CodeGenOpt::Aggressive).setErrorStr(&err).create();	
		if ( !EE ) {
			std::cerr << "Error making ExecutionEngine: " << err << std::endl;
			exit(-1);
		}
	}

	void* f0 = EE->getPointerToFunction(compiledModule->getFunction("create"));
	void* f1 = EE->getPointerToFunction(compiledModule->getFunction("fetch"));
	printf("thread %d got create at %p, fetch at %p\n", long(arg), f0, f1);
	delete EE;
	return NULL;
}

int main(int argc, char** argv)
{
	llvm::InitializeNativeTarget();

	if ( !llvm::llvm_is_multithreaded() ) llvm::llvm_start_multithreaded();
	pthread_t threads[THREADS];

	for ( unsigned i = 0; i < THREADS; ++i ) {
		if ( pthread_create(&threads[i], NULL, &go, (void*)i) ) {
			std::cerr << "Could not create thread " << i << std::endl;
			return -1;
		}
	}

	for ( unsigned i = 0; i < THREADS; ++i ) {
		if ( pthread_join(threads[i], NULL) ) {
			std::cerr << "Could not join thread " << i << std::endl;
			return -1;
		}
	}
	return 0;
}
